/* A syntactic analyser for miniML */

/* This parser parses strings of the form "a : tau" where a is a
   miniML expression where the binders have been explicitely annotated
   with their type (as in Pascal or C), and tau is the type of the
   expression.  It returns a pair containing (a,tau).  The expression
   a returned has been appropriately alpha-converted, and all binders
   are distinct.  No type-checking is done.

   Some examples of syntactically corrected expressions:

   3 : int

   let (x:bool) = true in (x,false) : bool * bool

   fun (x:int) -> x : int -> int

   (fun (x:int) -> let (h:int) = 7 in 
        fun (y:int) -> fun (z:int) -> (x,h)) 1 2 3 : int
*/

%{
open Ast

let gensym =
  let count = ref 0 in
  fun () -> count := !count+1; "x_"^(string_of_int !count)
%}

/* Tokens */
%token TRUE
%token FALSE
%token APOSTROPHE
%token STAR
%token SEMICOLON
%token COMMA
%token FUN
%token ARROW
%token LET
%token EQUAL
%token IN
%token <string> IDENT
%token <string> OP
%token <int> INT
%token LPAREN
%token RPAREN
%token EOF
%token FST
%token SND

%right prec_fun
%right prec_let
%right COMMA
%right ARROW
%left STAR

/* entry points */
%type <Ast.expression * Ast.exprtype> expression
%start expression

%%

expression: expr SEMICOLON type_expr EOF         { ($1 [],$3) }
;

expr:
    application                 { fun bind -> $1 bind }

  | expr COMMA expr             { fun bind -> Pair($1 bind, $3 bind) }

  | FUN LPAREN IDENT SEMICOLON type_expr RPAREN ARROW expr %prec prec_fun
       { fun bind -> 
	 let x = gensym() in
	 Fun(x, $5, $8 (($3,x)::bind)) }

  | LET LPAREN IDENT SEMICOLON type_expr RPAREN EQUAL expr IN expr %prec prec_let
       { fun bind ->
	 let x = gensym() in
	 Let(x, $5, $8 bind, $10 (($3,x)::bind)) }
;

application:
    application closed_expr     { fun bind -> App($1 bind, $2 bind) }
  | FST closed_expr             { fun bind -> Fst ($2 bind) }
  | SND closed_expr             { fun bind -> Snd ($2 bind) }
  | closed_expr                 { fun bind -> $1 bind }
;

closed_expr:
    IDENT               
      { fun bind -> 
        try Var (List.assoc $1 bind) 
        with Not_found -> failwith "The term is not closed" }
  | OP                  { fun bind -> Op $1 }
  | INT                 { fun bind -> Const (CInt $1) }
  | TRUE                { fun bind -> Const (CBool true) }
  | FALSE               { fun bind -> Const (CBool false) }
  | LPAREN expr RPAREN  { fun bind -> $2 bind }
;

type_expr:
    IDENT                     { Type_base $1 }
  | APOSTROPHE IDENT          { Type_var $2 }
  | type_expr ARROW type_expr { Type_arrow ($1, $3) }
  | type_expr STAR type_expr  { Type_product ($1, $3) }
  | LPAREN type_expr RPAREN   { $2 }
