(* Programming exercise no. 2 *)
open Ast

(* type environment *)

type env = (string * exprtype) list

let (gen_env, last_env, prev_env) =
  let count = ref 0 in
  ( (fun () -> let c = !count in count := !count+1; "env_"^(string_of_int c)),
    (fun () -> "env_"^(string_of_int !count)),
    (fun () -> "env_"^(string_of_int (!count-1))))

let gensym =
  let count = ref 0 in
  fun () -> count := !count+1; "x_"^(string_of_int !count)

let gentsym =
  let count = ref 0 in
  fun () -> count := !count+1; "a_"^(string_of_int !count)

let rec in_env x gamma =
  match gamma with
  | []       -> false
  | (h,_)::t -> if x = h then true else (in_env x t)

let rec look_env x gamma =
  match gamma with
  | []       -> failwith "empty env"
  | (h,v)::t -> if x = h then v else (look_env x t)

(* intermediate abstract syntax: all subterms carry type annotations *)

type typed_expression = 
    { exp : typed_expression_int;
      typ : exprtype }
and typed_expression_int =
  | TVar of string
  | TConst of constant
  | TOp of string
  | TFun of string * exprtype * typed_expression
  | TApp of typed_expression * typed_expression
  | TPair of typed_expression * typed_expression
  | TFst of typed_expression
  | TSnd of typed_expression
  | TLet of string * exprtype * typed_expression * typed_expression

let annotate_expression =
  let rec annotate_expression_int a gamma = 
    match a with 
    | Var x -> 
	if (in_env x gamma) 
	then { exp = TVar x; typ = (look_env x gamma) }
	else failwith "open term"
    | Const (CInt v) ->
	{ exp = TConst (CInt v); typ = Type_base "int" }
    | Const (CBool v) ->
	{ exp = TConst (CBool v); typ = Type_base "bool" }
    | Op _ -> failwith "not implemented"
    | Fun (x,t,a1) ->
	let a1t = annotate_expression_int a1 ((x,t)::gamma) in
	{ exp = TFun (x,t,a1t); typ = Type_arrow (t,a1t.typ) }
    | App (a1,a2) ->
	let a1t = annotate_expression_int a1 gamma in
	let a2t = annotate_expression_int a2 gamma in
	( match a1t.typ with
	| Type_arrow (t2,t1) -> 
	    if (a2t.typ = t2) 
	    then { exp = TApp (a1t,a2t); typ = t1 }
	    else failwith "type error"
	| _ -> failwith "type error" )
    | Pair (a1,a2) ->
	let a1t = annotate_expression_int a1 gamma in
	let a2t = annotate_expression_int a2 gamma in
	{ exp = TPair (a1t,a2t); typ = Type_product (a1t.typ,a2t.typ) }
    | Fst a1 ->
	let a1t = annotate_expression_int a1 gamma in
	( match a1t.typ with
	| Type_product (t1,_) -> 
	    { exp = TFst a1t; typ = t1 }
	| _ -> failwith "type error" )
    | Snd a1 ->
	let a1t = annotate_expression_int a1 gamma in
	( match a1t.typ with
	| Type_product (_,t2) -> 
	    { exp = TSnd a1t; typ = t2 }
	| _ -> failwith "type error" )
    | Let (x,t,a1,a2) ->
	let a1t = annotate_expression_int a1 gamma in
	if a1t.typ = t then 
	  let a2t = annotate_expression_int a2 ((x,t)::gamma) in
	  { exp = TLet (x,t,a1t,a2t); typ = a2t.typ }
	else failwith "type error"
  in fun a -> annotate_expression_int a []

(* typed closure conversion *)

let rec translate_env gamma =
  let tr_item (x,t) = (x, TL_Var x) in
  let r = List.map tr_item gamma  in
  TL_Record r  

let rec type_env gamma = 
  let tr_item (x,t) = (x, t) in
  let r = List.map tr_item gamma in
  TL_Type_record r

let rec translate_type tau =
  match tau with
  | Type_base t -> TL_Type_base t
  | Type_var a -> TL_Type_var a
  | Type_arrow (t2,t1) ->
      let alpha = gentsym() in
      TL_Type_exists 
	(alpha, TL_Type_product 
	   (TL_Type_fun 
	      (TL_Type_var alpha, translate_type t2, translate_type t1), 
	    TL_Type_var alpha))
  | Type_product (t1,t2) -> 
      TL_Type_product (translate_type t1, translate_type t2)

let translate =
  let rec translate_int gamma1 gamma2 a_source tau = 
  match a_source with
  | TVar x ->
      if (in_env x gamma1) 
      then TL_Proj ((TL_Var (last_env())), x)
      else TL_Var x
  | TConst i -> TL_Const i
  | TOp _ -> failwith "not implemented"
  | TFun (x,t,a) ->
      let gamma = List.append gamma2 gamma1 in
      let a_env = translate_env gamma in
      let a' = translate_int gamma [(x, translate_type t)] a.exp a.typ in
      let t_env = type_env gamma in
      TL_Pack (t_env, TL_Pair (TL_Fun (gen_env(), x, a'), a_env))
  | TApp (a1, a2) ->
      let x = gensym() in
      let a'_1 = translate_int gamma1 gamma2 a1.exp a1.typ in
      let a'_2 = translate_int gamma1 gamma2 a2.exp a2.typ  in
      TL_Open (a'_1, x, 
               TL_Call ((TL_Fst (TL_Var x)), (TL_Snd (TL_Var x)), a'_2))
  | TPair (a1, a2) ->
      let a'_1 = translate_int gamma1 gamma2 a1.exp a1.typ in
      let a'_2 = translate_int gamma1 gamma2 a2.exp a2.typ in
      TL_Pair (a'_1, a'_2)
  | TFst a -> 
      let a' = translate_int gamma1 gamma2 a.exp a.typ in
      TL_Fst a'
  | TSnd a -> 
      let a' = translate_int gamma1 gamma2 a.exp a.typ in
      TL_Snd a'
  | TLet (x, t, a1, a2) ->
      let a'_1 = translate_int gamma1 gamma2 a1.exp a1.typ in
      let a'_2 = translate_int gamma1 
	  ((x, translate_type t)::gamma2) a2.exp a2.typ in
      TL_Let (x, a'_1, a'_2)
  in      
  fun gamma1 gamma2 a tau ->
    let aa = annotate_expression a in
    if aa.typ = tau 
    then (translate_int gamma1 gamma2 aa.exp aa.typ, translate_type aa.typ)
    else failwith "type error"




