(* utility functions to manipulate miniML and target-language terms *)

open Ast

(* parse_expression s  parses the string s and return a miniML term in
   which all bindings are distincts *)
let parse_expression s =
  (Parser.expression Lexer.token (Lexing.from_string s) 
     : expression * exprtype)

(* print_type t  returns a string containing a pretty-printed
   representation of the miniML type t *)
let rec print_type t =
  match t with
  | Type_base s          -> s
  | Type_var v           -> "'"^v
  | Type_arrow (t1,t2)   -> "("^(print_type t1)^" -> "^(print_type t2)^")"
  | Type_product (t1,t2) -> "("^(print_type t1)^" * "^(print_type t2)^")"

(* print_expression a  returns a string containing a pretty-printed
   representation of the expression a *)
let rec print_expression a =
  match a with
  | Var v         -> v
  | Const (CInt i) -> string_of_int i
  | Const (CBool b) -> string_of_bool b 
  | Op o          -> o
  | Fun (s,t,a1)  -> "fun ("^s^":"^(print_type t)^") -> "^(print_expression a1)
  | App (a1,a2)   -> 
      "(("^(print_expression a1)^") ("^(print_expression a2)^"))"
  | Pair (a1,a2)  -> "("^(print_expression a1)^", "^(print_expression a2)^")"
  | Fst (a1)       -> "(fst ("^(print_expression a1)^"))"
  | Snd (a1)       -> "(snd ("^(print_expression a1)^"))"
  | Let (v,t,a1,a2) -> 
      "let ("^v^":"^(print_type t)^") = "^
      (print_expression a1)^" in "^(print_expression a2)

(* print_tl_type t  returns a string containing a pretty-printed
   representation of the target_language type t *)
let rec print_tl_type t =
  match t with
  | TL_Type_base s          -> s
  | TL_Type_var v           -> "'"^v
  | TL_Type_fun (t1,t2,t3)  -> 
      "(("^(print_tl_type t1)^
      " * "^(print_tl_type t2)^") -> "^(print_tl_type t3)^")"
  | TL_Type_product (t1,t2) -> 
      "("^(print_tl_type t1)^" * "^(print_tl_type t2)^")"
  | TL_Type_record l        -> 
      let print_item (e,t) b = e^":"^(print_tl_type t)^"; "^b
      in "{ "^(List.fold_right print_item l "")^" }"
  | TL_Type_exists (s,t)    -> 
      "exists '"^s^"."^(print_tl_type t)

(* print_tl_expression a  returns a string containing a pretty-printed
   representation of the target_language expression a *)
let rec print_tl_expression a = 
  match a with
  | TL_Const (CInt i) -> string_of_int i
  | TL_Const (CBool b) -> string_of_bool b
  | TL_Var x         -> x
  | TL_Op s          -> s
  | TL_Fun (x1,x2,a) -> "fun ("^x1^", "^x2^") -> "^(print_tl_expression a)
  | TL_Call (a,a1,a2) -> 
      "("^(print_tl_expression a)^" ("^(print_tl_expression a1)^
      ", "^(print_tl_expression a2)^"))"
  | TL_Pair (a1,a2)  -> 
      "("^(print_tl_expression a1)^", "^(print_tl_expression a2)^")"
  | TL_Fst (a)       -> "(fst ("^(print_tl_expression a)^"))"
  | TL_Snd (a)       -> "(snd ("^(print_tl_expression a)^"))"
  | TL_Record l      ->
      let print_item (e,a) b = e^"="^(print_tl_expression a)^"; "^b
      in "{ "^(List.fold_right  print_item l "")^" }"
  | TL_Proj (a,e)    -> "(("^(print_tl_expression a)^")."^e^")"
  | TL_Pack (t,a)    -> 
      "(pack "^(print_tl_type t)^" with "^(print_tl_expression a)^")"
  | TL_Open (a1,x,a2) ->
      "(open "^(print_tl_expression a1)^" as "^x^" in "^
      (print_tl_expression a2)^")"
  | TL_Let (v,a1,a2) ->
      "let "^v^" = "^(print_tl_expression a1)^" in "^(print_tl_expression a2)

